from typing import Callable, Dict, List
import pathlib
import json
import importlib
import pkg_resources

from excal import visitor


class ModuleInterface:
    """Represents a plugin interface. A plugin has a single register function."""

    @staticmethod
    def register(PluginManager) -> None:
        """Register the necessary items in the game character factory."""


def import_module(name: str) -> ModuleInterface:
    """Imports a module given a name."""
    return importlib.import_module(name)  # type: ignore


class PluginManager():
    """Class to find, register and load Plugins."""
    def __init__(self):
        self.plugins: Dict[str, "visitor.NodeVisitor"] = {}

    def register(self, plugin_name: str, plugin_init_fun: Callable[[], "NodeVisitor"]) -> None:
        try:
            plugin: "visitor.NodeVisitor" = plugin_init_fun()
            self.plugins[plugin_name] = plugin
        except Exception:
            print(f"error loading Plugin {plugin_name}")

    def getPluginList(self) -> List["visitor.NodeVisitor"]:
        retList: List[visitor.NodeVisitor] = []
        for key, val in self.plugins.items():
            retList.append(val)
        return retList

    def loadPlugins(self):
        try:
            plugin_file = open(f"{pathlib.Path(__file__).resolve().parent.parent.parent}/plugins.json")
            plugin_data = json.load(plugin_file)["plugins"]
            for plugin_file in plugin_data:
                # print(f"plugin {plugin_file} loading")
                plugin = import_module(plugin_file)
                plugin.register(self)
        except Exception:
            pass

        for entry_point in pkg_resources.iter_entry_points('excal_plugins'):
            # print(f"plugin {entry_point.name} loading")
            entry_point.load()(self)



from pathlib import Path
from typing import List
from clang.cindex import Index as CIndex, Cursor, TranslationUnit

from excal.astNode import AstNode, Location, Token


class CWalker:
    """Create AST and translate it to astNodes"""
    def __init__(self, c_file: Path, clang_args: List[str]) -> None:
        self.anonymous_counter = 0
        self.path: Path = c_file
        # use clang to parse C file
        self.index: CIndex = CIndex.create()
        self.parsed_unit: TranslationUnit = self.index.parse(c_file, args=clang_args)
        self.root_node: Cursor = self.parsed_unit.cursor
        self.ast: AstNode

    def walkRec(self, node: Cursor, indent: str, ast: AstNode) -> None:
        indent += '  '
        for child_node in node.get_children():
            try:
                if not self.path.samefile(child_node.location.file.name):
                    continue
            except Exception:
                continue
            ast_child = AstNode(child_node.kind, child_node.location.file.name,
                                child_node.location.line, child_node.location.column,
                                node.extent.end.line, node.extent.end.column,
                                str(child_node.spelling), ast.indent_level + 1,
                                str(child_node.type.spelling), ast,
                                [Token(x.kind, x.spelling, Location(x.location.line, x.location.column)) for x in child_node.get_tokens()])


            ast.add_child(ast_child)
            self.walkRec(child_node, indent, ast_child)

    def walk(self) -> AstNode:
        self.ast = AstNode(self.root_node.kind, "", 0, 0, self.root_node.extent.end.line, self.root_node.extent.end.column, str(self.root_node.spelling), 0, "", None, [Token(x.kind, x.spelling, Location(x.location.line, x.location.column)) for x in self.root_node.get_tokens()])
        # for token in self.root_node.get_tokens():
        # print(token.spelling)
        # print(f"{token.location.line}: {token.location.column}: {token.kind}")
        self.walkRec(self.root_node, '', self.ast)
        return self.ast



"""EXCAL - Extendable Clang AST based Linter."""
import os
import glob
from typing import List
import configparser
from argparse import ArgumentParser

from excal.analyzer import Analyzer
from excal.pluginManager import PluginManager
from excal.output import Output


def readConfig(args):
    if len(args.files) != 1:
        return args

    path: str = args.files[0]

    if not os.path.isdir(path):
        return args
    if not path.endswith('/') and not path.endswith('\\'):
        path += '/'
    if not os.path.exists(f"{path}setup.cfg"):
        return args

    config = configparser.ConfigParser()
    config.read(f"{path}setup.cfg")

    if 'EXCAL' not in config.sections():
        return args

    cfg = config['EXCAL']
    for key in cfg:
        if key == "force_cpp":
            args.force_cpp = cfg[key] == 'True'
        if key == "includes":
            args.include.extend([l.strip() for l in cfg[key].split('\n')])
        if key == "exclude_files":
            args.exclude_files.extend(
                [f'{path}{line.strip()}'
                 for line in cfg[key].split('\n') if line != ''])
    return args


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--args", "-a", nargs="*", required=False,
                        default=[], help="Additional clang arguments")
    parser.add_argument("--argfile", type=str, default=None, required=False,
                        help="File containing clang arguments")
    parser.add_argument("--include", "-i", nargs="+", required=False, default=[],
                        help="include directories")
    parser.add_argument("--files", "-f", nargs="+", required=True,
                        help="File to translate")
    parser.add_argument("--exclude-files", "-e", nargs="+", required=False, default=[],
                        help="files within files directory, that should be ignored")
    parser.add_argument("--print-tree", "-p", nargs='?', required=False,
                        const=True, default=False, help="print AST of input files")
    parser.add_argument("--force-cpp", "-cpp", nargs='?', required=False,
                        const=True, default=False,
                        help="forces files to be analyzed as C++ files.")
    parser.add_argument("--extensions", "-ext", nargs="+", required=False,
                        default=[], help="file extensions to parse.")
    parser.add_argument("--print-token-tree", "-ptt", nargs='?', required=False,
                        const=True, default=False, help="print AST plus corresponding Tokens")

    args = parser.parse_args()

    args = readConfig(args)

    clang_args = args.args

    if args.argfile is not None:
        with open(args.argfile, "r") as fl:
            lines = [ln.strip() for ln in fl.readlines()]
            clang_args.extend(lines)

    if args.force_cpp:
        clang_args.extend(['-x', 'c++'])

    if args.extensions != []:
        FILE_EXTENSIONS = args.extensions
    else:
        FILE_EXTENSIONS: List[str] = ["c", "h", "cpp", "hpp"]

    exfiles = set()
    for ex_file in args.exclude_files:
        if not os.path.isfile(ex_file):
            for ext in FILE_EXTENSIONS:
                exfiles |= set(glob.glob(f"{ex_file}/**/*.{ext}", recursive=True))
        else:
            exfiles.add(ex_file)

    analyze_files = []
    for in_file in args.files:
        if os.path.isdir(in_file):
            for ext in FILE_EXTENSIONS:
                new_files = glob.glob(f"{in_file}/**/*.{ext}", recursive=True)
                analyze_files = analyze_files + [f for f in new_files if f not in exfiles]
        else:
            analyze_files = analyze_files + [in_file]

    pm: PluginManager = PluginManager()
    pm.loadPlugins()
    out: Output = Output()
    analyzer = Analyzer(args.include, clang_args, pm, out)

    printState = 0
    if(args.print_tree):
        printState = 1
    if(args.print_token_tree):
        printState = 2

    for in_file in analyze_files:
        analyzer.analyze(in_file, printState)

    out.printError()


if __name__ == "__main__":
    main()



import sys
from typing import List
from typing import Tuple


class Output():
    """Output class. Will be extended by more output options."""
    def __init__(self):
        self.exeptions: List[Tuple[str, int, int, str]] = []
        return

    def addExeptioins(self, ex: List[Tuple[str, int, int, str]]):
        self.exeptions.extend(ex)

    def printError(self):
        for e in self.exeptions:
            print(f'{e[0]}: {e[1]}:{e[2]} : {e[3]}', file=sys.stderr)



from pathlib import Path
from typing import List

from excal.astNode import AstNode
from excal.cwalker import CWalker
from excal.visitor import CXXAstNodeVisitor
from excal.pluginManager import PluginManager
from excal.output import Output


class Analyzer:
    """Analyzer Class, will prepare analysis, call clang to create the AST, translate the AST to astNodes, and call the
    visitor class to get all linter results."""
    def __init__(self, include_dirs: List[str], clang_args: List[str], pm: PluginManager, out: Output) -> None:
        self.include_dirs: List[Path] = [Path(p) for p in include_dirs]
        # Check for additional includes in args
        arg_includes = [arg[2:] for arg in clang_args if arg.startswith("-I")]
        self.include_dirs.extend([Path(inc) for inc in arg_includes if inc not in include_dirs])
        # And store only non includes
        self.additional_arguments = [arg for arg in clang_args if not arg.startswith("-I")]
        self.ast: AstNode
        self.pm: PluginManager = pm
        self.out: Output = out

    def find_file(self, file_name: str) -> Path:
        result = Path(file_name)
        if result.exists():
            return result
        for inc_dir in self.include_dirs:
            result = inc_dir / file_name
            if result.exists():
                return result
        return None

    def clang_args(self) -> List[str]:
        result = self.additional_arguments.copy()
        result.extend([f"-I{inc}" for inc in self.include_dirs])
        return result

    def printAstRec(self, node: AstNode, printState: int):
        print(node)
        if printState == 2:
            node.printTokens()
        for child in node.children:
            self.printAstRec(child, printState)

    def analyze(self, file_name: str, printAst: int) -> None:
        in_path = self.find_file(file_name)
        # print(f"Parsing {file_name}")
        walker = CWalker(in_path, self.clang_args())
        self.ast = walker.walk()

        visitor = CXXAstNodeVisitor(self.ast, self.pm)
        visitor.visit()
        self.out.addExeptioins(visitor.getExceptions())
        if printAst > 0:
            self.printAstRec(self.ast, printAst)
